<h2>Development</h2>

<p>Rodauth, while a mature framework, is still under active development.  You can join in on the discussions, ask questions, suggest features, and discuss Rodauth in general on <a href="https://github.com/jeremyevans/rodauth/discussions">GitHub Discussions</a> or by joining the <a href='http://groups.google.com/group/rodauth'>rodauth Google Group</a>.</p>

<h3>Reporting Bugs</h3>

<p>To report a bug in Rodauth, use <a href="https://github.com/jeremyevans/rodauth/issues">GitHub Issues</a>.  If you aren't sure if something is a bug, post a question on <a href="https://github.com/jeremyevans/rodauth/discussions">GitHub Discussions</a> or the <a href='http://groups.google.com/group/rodauth'>Google Group</a>.  Note that GitHub Issues should not be used to ask questions about how to use Rodauth; use GitHub Discussions or the Google Group for that.</p>

<h3>Source Code</h3>

<p>The master source code repository is <a href="https://github.com/jeremyevans/rodauth">jeremyevans/rodauth on GitHub</a>.</p>

<h3>Submitting Patches</h3>

<p>The easiest way to contribute is to use Git, post the changes to a public repository, and send a pull request, either via GitHub, or the Google Group.  Posting patches to the bug tracker or the Google Group works fine as well.</p>

<h3>License</h3>

<p>Rodauth is distributed under the <a href="rdoc/files/MIT-LICENSE.html">MIT License</a>.  Patches are assumed to be submitted under the same license as Rodauth.</p>

